"""Collection of utilities for testing ros nodes less painfully.
"""

import pyrostest.rostest_utils

from pyrostest.launch_tools import with_launch_file, launch_node
from pyrostest.ros_test import RosTest




"""Utilities for launching ros nodes for tests.
"""
import functools
import os
import sys
import time

import roslaunch

import pyrostest.rostest_utils

class RosLaunchException(Exception):
    """An Exception thrown when the with_launch_file decorator is misused.
    """
    pass



class ROSLauncher(roslaunch.scriptapi.ROSLaunch):
    """
    ROSLaunch but allows the use of launch files.

    This was found by peering into the roslaunch util source code from
    `which roslaunch`.
    """
    def __init__(self, files, port=11311):
        super(ROSLauncher, self).__init__()
        uuid = roslaunch.rlutil.get_or_generate_uuid(None, True)
        self.parent = roslaunch.parent.ROSLaunchParent(uuid,
                files, is_core=False, port=port)


# The following two methods are intended for use together. You need to be
# aware of the following limitations:
# `with_launch_file` should always be called first. That is:
#
# ```
# @with_launch_file(*args)
# @launch_node(*node_args)
# def test_function(self):
#     ...
# ```

# is valid, as is
#
# ```
# @launch_node(*node_args)
# @launch_node(*other_node_args)
# def test_function(self):
#     ...
# ```
# however, neither of these would be valid:
#
# ```
# @launch_node(*node_args)
# @with_launch_file(*args)
# def test_function(self):
#     ...
# ```
#
# ```
# @with_launch_file(*args)
# @launch_node(*node_args)
# @with_launch_file(*more_launch_args)
# def test_function(self):
#     ...
# ```
#
# this is handled by having `with_launch_file` check for a running launcher for
# its rosmaster port, and if one exists already, raising an error. Otherwise it
# just works.
# Similarly, launch_node will use a launcher if one exists, otherwise the
# outermost launch_node will create a master launcher and clean up after itself
# and all child nodes launched for the test.

def with_launch_file(package, launch, **kwargs):
    """Decorator to source a launch file for running nodes.

    This should always be run first.

    This and launch nodes work together gracefully, as long as you follow the
    guidelines outlined in `buzzmobile/tests/test_utils/launch_tools.py`.
    """
    def launcher(func):
        """Decorator function created by the decorator-gen.
        """
        
        @functools.wraps(func)
        def new_test(self):
            """Wrapper around the user provided test that runs a launch file.
            """
            # This should be in new_test and not outside because moving it
            # outside causes errors at test collection time instead of test run
            # time, at the cost of a minor decrease in performance.
            try:
                full_name = roslaunch.rlutil.resolve_launch_arguments([package,
                    launch])
            except roslaunch.core.RLException:
                raise Exception('The package {}/{} does not exist. Make sure'
                        'the name is correct and you have initialized your ROS'
                        'environment'.format(package, launch))
            # set env variables and add argvs to sys.argv
            os.environ['ROS_MASTER_URI'] = self.rosmaster_uri
            new_argvs = ['{}:={}'.format(k, v) for k, v in kwargs.iteritems()]
            sys.argv.extend(new_argvs)

            ros_launcher = ROSLauncher(full_name, port=self.port)
            ros_launcher.start()
            if self.port in self.LAUNCHER:
                raise RosLaunchException('Rosmaster port {} already in use. '
                'You must call use @with_launch_file only once for any single '
                'test, and before any @launch_node calls.'.format(self.port))

            self.LAUNCHER[self.port] = ros_launcher

            try:
                temp = func(self)
            except Exception as exc:
                raise exc
            finally:
                self.LAUNCHER[self.port].stop()
                # clean argvs from sys.argv
                sys.argv = sys.argv[:len(new_argvs)]
                del self.LAUNCHER[self.port]
            return temp
        return new_test
    return launcher

def launch_node(package, name, namespace=None):
    """Decorator to manage running a node and shutting it down gracefully.

    Note that this will wrap itself up cleanly and launch all nodes with a
    single launcher, instead of multiples.
    """
    if not namespace:
        namespace = '/'+package
    def launcher(func):
        """Actual decorator generated by the above.
        """
        @functools.wraps(func)
        def new_test(self):
            """Wrapper around the user-provided test that runs a ros node.
            """
            is_master = False
            if self.port not in self.LAUNCHER:
                # set env variables and add argvs to sys.argv
                os.environ['ROS_MASTER_URI'] = self.rosmaster_uri
                launch = ROSLauncher([], port=self.port)
                launch.start()
                self.LAUNCHER[self.port] = launch
                is_master = True
            else:
                launch = self.LAUNCHER[self.port]

            env = {'ROS_MASTER_URI': self.rosmaster_uri}
            node = roslaunch.core.Node(package, name, namespace=namespace,
                    env_args=env.iteritems())

            process = launch.launch(node)
            # Beware this is a bit of a hack, and will currently not work if we
            # want to run more than 1 node with the same name.
            while not any(nn.split('/')[-1].startswith(name.replace('.', '_'))
                    for nn in
                    pyrostest.rostest_utils.my_get_node_names(uri=self.rosmaster_uri)):
                time.sleep(.1)
            try:
                temp = func(self)
            except:
                raise
            finally:
                process.stop()
            if is_master:
                self.LAUNCHER[self.port].stop()
                del self.LAUNCHER[self.port]
            return temp

        return new_test
    return launcher



#! /usr/bin/env python

"""A rosnode in a subprocess that will publish input from stdin.

This takes in a pickled tuple of three values: ROSTOPIC, ROSMSG_TYPE, and
QUEUE_SIZE as its only command line argument. This means that it must be able
to unpickle the ROSMSG_TYPE argument, which requires importing std_msgs or
similar packages. So rosinit must have been done on this thread. This will have
been done by default, but might be a cause for odd issues.
"""

import cPickle as pickle
import sys
from StringIO import StringIO

import rospy


def publish(topic, msg_type):
    """A function to isolate the publishing logic.
    """
    rospy.init_node('mock_publish_'+topic.split('/')[-1], anonymous=True,
            disable_signals=True)
    pub = rospy.Publisher(topic, msg_type, queue_size=QUEUE_SIZE)
    msg = msg_type()
    sio = StringIO()
    msg.serialize(sio)
    while 1:
        data = sys.stdin.read(sio.len)
        msg.deserialize(data)
        pub.publish(msg)


if __name__ == '__main__':
    ROSTOPIC, ROSMSG_TYPE, QUEUE_SIZE = pickle.loads(sys.argv[-1])
    publish(ROSTOPIC, ROSMSG_TYPE)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import codecs
import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

required = [
    'psutil==5.1.3'
    ]

test = [
    'pytest'
    ]

setup(
    name='pyrostest',
    version='0.1.9',
    description='The most lit ros testing framework',
    long_description=long_description,
    packages=['pyrostest'],
    package_data={
        'pyrostest': ['data/*']
    },
    install_requires=required,
    tests_require=test,
    url='https://github.com/gtagency/pyrostest',
    author='Joshua Morton',
    author_email='joshua.morton13@gmail.com',
    maintainer='Raphael Gontijo Lopes',
    maintainer_email='raphaelgontijolopes@gmail.com',
    license='MIT',
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        ],
    )





"""A collection of utilities to make testing with ros less painful.
"""

import os
import functools
import random
import socket
import subprocess
import time

import psutil
import rosgraph
import rosnode


def my_get_node_names(namespace=None, uri='http://localhost:11311'):
    """Monkeypatches get_node_names with a non-default ROS_MASTER_URI.
    """
    old_master = rosgraph.Master
    rosgraph.Master = functools.partial(rosgraph.Master, master_uri=uri)
    nodenames = rosnode.get_node_names(namespace=namespace)
    rosgraph.Master = old_master
    return nodenames


def rand_port():
    """Picks a random port number.

    This is potentially unsafe, but shouldn't generally be a problem.
    """
    return random.randint(10311, 12311)


class RosTestMeta(type):
    """Metaclass for RosTest that adds the setup/teardown we want.
    """
    def __new__(mcs, name, bases, dct):

        # It will break unless we throw in fake setup and teardown methods if
        # the real ones don't exist yet.

        def noop(_):
            """Do nothing function.

            This is injected if there is no user-defined setUp or tearDown
            method on an instance of RosTest.
            """
            pass

        try:
            old_setup = dct['setUp']
        except KeyError:
            old_setup = noop
        try:
            old_teardown = dct['tearDown']
        except KeyError:
            old_teardown = noop

        def new_setup(self):
            """Wrapper around the user-defined setUp method that runs roscore.
            """
            self.port = rand_port()
            self.rosmaster_uri = 'http://{}:{}'.format(socket.gethostname(),
                    self.port)
            env = {k:v for k, v in os.environ.iteritems()}
            env.update({'ROS_MASTER_URI': self.rosmaster_uri})
            roscore_initialized = False
            while not roscore_initialized:
                self.roscore = subprocess.Popen(
                        ['roscore', '-p', str(self.port)], env=env)
                # an error will return a nonzero errorcode, and None indicates
                # that the process is still running, so falsy results are good
                time.sleep(1)
                if not self.roscore.poll():
                    roscore_initialized = True
                else:
                    self.roscore.kill()
                    self.roscore = None
            old_setup(self)

        def new_teardown(self):
            """Wrapper around the user-defined tearDown method to end roscore.
            """
            proc = psutil.Process(self.roscore.pid)
            children = proc.children(recursive=True)

            old_teardown(self)
            self.roscore.kill()
            self.roscore.wait()
            self.roscore = None
            for child in children:
                try:
                    child.terminate()
                    child.wait()
                except psutil.NoSuchProcess:
                    # its possible the process has already been killed
                    pass


        dct['setUp'] = new_setup
        dct['tearDown'] = new_teardown
        dct['setUp'].__name__ = 'setUp'
        dct['tearDown'].__name__ = 'tearDown'
        return super(RosTestMeta, mcs).__new__(mcs, name, bases, dct)




#! /usr/bin/env python

"""A ros node that will subscribe to a single message and print it to stdout.

It takes in pickled configuration string that should be the ROSTOPIC and
ROSMSG_TYPE pickled on the other end. This also means that this node must be
run from a context where it can import any of the buzzmobile ros messages.

It prints the recieved data to stdout serialized for use by the main thread.
"""
import cPickle as pickle
import sys

import rospy



def subscribe(topic, msg_type):
    """A function to isolate some of the subscriber logic.
    """
    rospy.init_node('mock_subscribe_'+topic.split('/')[-1], anonymous=True,
            disable_signals=True)
    def listener_cb(message):
        """A default callback for our subscriber.
        """
        message.serialize(sys.stdout)
        rospy.signal_shutdown('message sent')

    rospy.Subscriber(topic, msg_type, listener_cb)
    rospy.spin()



if __name__ == '__main__':
    ROSTOPIC, ROSMSG_TYPE = pickle.loads(sys.argv[-1])
    subscribe(ROSTOPIC, ROSMSG_TYPE)



"""Utilities that deal with creating fake rosnodes for testing.

Contains tools to send and recieve data from fake nodes and topics. This is
needed because ROS doesn't allow you to run multiple nodes from the same
process, so instead we have to set up a background process and communicate with
it over the system.
"""

import contextlib
import cPickle as pickle
import os
import pkg_resources
import subprocess
import time
import threading
from StringIO import StringIO
import unittest

import pyrostest.rostest_utils

FileNotFoundError = IOError


class TimeoutError(Exception):
    """Py3 shim, represents no response from a syscall.
    """
    pass


class NoMessage(Exception):
    """Exception for a lack of message from a Node.
    """
    pass


class MockPublisher(object):
    """Mock of a node object for testing.
    """
    def __init__(self, topic, msg_type, queue_size):
        self.topic = topic
        self.msg_type = msg_type
        pub_data = pickle.dumps((topic, msg_type, queue_size))
        # dynamically looks up the location of the publisher.py file in
        # relation to this file (they should be in the same dir)
        location = pkg_resources.resource_filename(__name__, "publisher.py")
        if not os.path.isfile(location):
            raise FileNotFoundError('{} cannot be located'.format(location))

        self.proc = subprocess.Popen(['python', location, pub_data],
                stdin=subprocess.PIPE)

    def send(self, value):
        """Sends data to be publoshed to the mocked topic.
        """
        time.sleep(.5)
        value.serialize(self.proc.stdin)

    def kill(self):
        """Kills the publisher ros node process.
        """
        self.proc.kill()
        self.proc.wait()


class MockSubscriber(object):
    """Wrapper around a node used for testing.
    """

    def __init__(self, topic, msg_type, timeout):
        self.timeout = timeout
        self.topic = topic
        self.msg_type = msg_type
        self.killed = False

        location = pkg_resources.resource_filename(__name__, "subscriber.py")
        if not os.path.isfile(location):
            raise FileNotFoundError('{} cannot be located'.format(location))
        self.proc = subprocess.Popen(['python', location,
            pickle.dumps((topic, msg_type))], stdout=subprocess.PIPE)
        self._message = None

    def kill(self):
        """Kills the subscriber ros node process.
        """
        self.proc.kill()
        self.proc.wait()

    @property
    def message(self):
        """Getter for the message property.

        Makes sure you've actually recieved a message before providing one.
        """
        if not self._message:
            msg = self.msg_type()
            sio = StringIO()
            msg.serialize(sio)
            # This weird business with StringIO is to make sure we read the
            # correct number of bytes. Essentially, a serialized ros message
            # is a fixed length string sent over the channel, so we serialize
            # the default message of a type and then read that number of bytes
            # so that we don't encounter any issues with reading too much or
            # too little data.

            def kill_proc():
                """Kills the proccess and sets a flag to raise an error.
                """
                self.proc.kill()
                self.killed = True

            timer = threading.Timer(self.timeout, kill_proc)
            try:
                timer.start()
                data = self.proc.stdout.read(sio.len)
            finally:
                timer.cancel()

            if self.killed:
                raise TimeoutError('No message published to {} within {} '
                        'seconds. Needed {} bytes, got {}.'.format(self.topic,
                            self.timeout, sio.len, len(data)))
            msg.deserialize(data)
            self._message = msg
        return self._message


def _await_node(topic, prefix, rosmaster_uri):
    """Helper to block until a node is availible in ROS.
    """
    no_ns = topic.split('/')[-1]
    while not any(nn.split('/')[-1].startswith(''.join([prefix, '_', no_ns]))
            for nn in
            pyrostest.rostest_utils.my_get_node_names(uri=rosmaster_uri)):
        time.sleep(.1)


class RosTest(unittest.TestCase):
    """A subclass of TestCase that exposes some additional ros-related attrs.

    self.port is the port this instance will run on.
    self.rosmaster_uri is equivalent to the ROS_MASTER_URI environmental var
    """
    __metaclass__ = pyrostest.rostest_utils.RosTestMeta

    def __init__(self, *args, **kwargs):
        super(RosTest, self).__init__(*args, **kwargs)
        self.LAUNCHER = dict()

    @contextlib.contextmanager
    def check_topic(self, topic, rosmsg_type, timeout=10):
        """Context manager that monitors a rostopic and gets a message sent to
        it.
        """
        test_node = MockSubscriber(topic, rosmsg_type, timeout=timeout)
        _await_node(topic, 'mock_subscribe', self.rosmaster_uri)

        try:
            yield test_node
        finally:
            test_node.kill()

    @contextlib.contextmanager
    def mock_pub(self, topic, rosmsg_type, queue_size=1):
        """Mocks a node and cleans it up when done.
        """
        pub = MockPublisher(topic, rosmsg_type, queue_size)
        _await_node(topic, 'mock_publish', self.rosmaster_uri)
        try:
            yield pub
        finally:
            pub.kill()

